@@ -17,7 +17,7 @@ import environment
 import dependencies
 import mlog
 import copy, os
-
+from mesonlib import File
 
 known_basic_kwargs = {'install' : True,
                       'c_pch' : True,
@@ -129,13 +129,11 @@ class Build:
         return self.external_link_args.get(compiler.get_language(), [])
 
 class IncludeDirs():
-    def __init__(self, curdir, dirs, kwargs):
+    def __init__(self, curdir, dirs):
         self.curdir = curdir
         self.incdirs = dirs
         # Interpreter has validated that all given directories
         # actually exist.
-        if len(kwargs) > 0:
-            raise InvalidArguments('Includedirs function does not take keyword arguments.')
 
     def get_curdir(self):
         return self.curdir
@@ -207,7 +205,7 @@ class BuildTarget():
             # Holder unpacking. Ugly.
             if hasattr(s, 'held_object'):
                 s = s.held_object
-            if isinstance(s, str):
+            if isinstance(s, File):
                 if not s in added_sources:
                     self.sources.append(s)
                     added_sources[s] = True
@@ -218,7 +216,10 @@ class BuildTarget():
 
     def validate_sources(self):
         if len(self.sources) > 0:
-            first = os.path.split(self.sources[0])[1]
+            firstname = self.sources[0]
+            if isinstance(firstname, File):
+                firstname = firstname.fname
+            first = os.path.split(firstname)[1]
             (base, suffix) = os.path.splitext(first)
             if suffix == '.rs':
                 if self.name != base:
@@ -248,14 +249,18 @@ class BuildTarget():
         if 'link_with' in self.kwargs:
             self.kwargs['link_with'] = self.unpack_holder(self.kwargs['link_with'])
 
-    def extract_objects(self, srclist):
+    def extract_objects(self, srcargs):
         obj_src = []
-        for src in srclist:
-            if not isinstance(src, str):
-                raise coredata.MesonException('Extraction arguments must be strings.')
-            if src not in self.sources:
-                raise coredata.MesonException('Tried to extract unknown source %s.' % src)
-            obj_src.append(src)
+        for srclist in srcargs:
+            if not isinstance(srclist, list):
+                srclist = [srclist]
+            for src in srclist:
+                if not isinstance(src, str):
+                    raise coredata.MesonException('Extraction arguments must be strings.')
+                src = File(False, self.subdir, src)
+                if src not in self.sources:
+                    raise coredata.MesonException('Tried to extract unknown source %s.' % src)
+                obj_src.append(src)
         return ExtractedObjects(self, obj_src)
 
     def get_rpaths(self):
@@ -552,6 +557,7 @@ class GeneratedList():
         self.infilelist = []
         self.outfilelist = []
         self.outmap = {}
+        self.extra_depends = []
 
     def add_file(self, newfile):
         self.infilelist.append(newfile)
@@ -665,11 +671,14 @@ class CustomTarget:
                     'command' : True,
                     'install' : True,
                     'install_dir' : True,
-                    'build_always' : True}
+                    'build_always' : True,
+                    'depends' : True}
+
     def __init__(self, name, subdir, kwargs):
         self.name = name
         self.subdir = subdir
         self.dependencies = []
+        self.extra_depends = []
         self.process_kwargs(kwargs)
         self.extra_files = []
         self.install_rpath = ''
@@ -741,6 +750,15 @@ class CustomTarget:
         self.build_always = kwargs.get('build_always', False)
         if not isinstance(self.build_always, bool):
             raise InvalidArguments('Argument build_always must be a boolean.')
+        extra_deps = kwargs.get('depends', [])
+        if not isinstance(extra_deps, list):
+            extra_deps = [extra_deps]
+        for ed in extra_deps:
+            while hasattr(ed, 'held_object'):
+                ed = ed.held_object
+            if not isinstance(ed, CustomTarget) and not isinstance(ed, BuildTarget):
+                raise InvalidArguments('Can only depend on toplevel targets.')
+            self.extra_depends.append(ed)
 
     def get_basename(self):
         return self.name
@@ -785,6 +803,9 @@ class RunTarget:
     def get_generated_sources(self):
         return []
 
+    def get_sources(self):
+        return []
+
     def get_subdir(self):
         return self.subdir
 
