@@ -21,21 +21,27 @@ from coredata import MesonException
 about. To support a new compiler, add its information below.
 Also add corresponding autodetection code in environment.py."""
 
-header_suffixes = ['h', 'hh', 'hpp', 'hxx', 'H', 'moc']
+header_suffixes = ['h', 'hh', 'hpp', 'hxx', 'H', 'moc', 'vapi']
 cpp_suffixes = ['cc', 'cpp', 'cxx', 'h', 'hh', 'hpp', 'hxx', 'c++']
 c_suffixes = ['c']
 clike_suffixes = c_suffixes + cpp_suffixes
 obj_suffixes = ['o', 'obj']
 
 def is_header(fname):
+    if hasattr(fname, 'fname'):
+        fname = fname.fname
     suffix = fname.split('.')[-1]
     return suffix in header_suffixes
 
 def is_source(fname):
+    if hasattr(fname, 'fname'):
+        fname = fname.fname
     suffix = fname.split('.')[-1]
     return suffix in clike_suffixes
 
 def is_object(fname):
+    if hasattr(fname, 'fname'):
+        fname = fname.fname
     suffix = fname.split('.')[-1]
     return suffix in obj_suffixes
 
@@ -395,6 +401,14 @@ void bar() {
 '''
         return self.compiles(templ % (prefix, typename, membername))
 
+    def has_type(self, typename, prefix):
+        templ = '''%s
+void bar() {
+    sizeof(%s);
+};
+'''
+        return self.compiles(templ % (prefix, typename))
+
     def thread_flags(self):
         return ['-pthread']
 
@@ -795,20 +809,19 @@ class ValaCompiler():
 
     def sanity_check(self, work_dir):
         src = 'valatest.vala'
-        obj = 'valatest.c'
         source_name = os.path.join(work_dir, src)
         ofile = open(source_name, 'w')
         ofile.write('''class SanityCheck : Object {
 }
 ''')
         ofile.close()
-        pc = subprocess.Popen(self.exelist + ['-C', '-o', obj, src], cwd=work_dir)
+        pc = subprocess.Popen(self.exelist + ['-C', '-c', src], cwd=work_dir)
         pc.wait()
         if pc.returncode != 0:
             raise EnvironmentException('Vala compiler %s can not compile programs.' % self.name_string())
 
     def can_compile(self, fname):
-        return fname.endswith('.vala')
+        return fname.endswith('.vala') or fname.endswith('.vapi')
 
 class RustCompiler():
     def __init__(self, exelist, version):
@@ -1003,12 +1016,18 @@ def get_gcc_soname_args(gcc_type, shlib_name, path, soversion):
         raise RuntimeError('Not impelented yet.')
 
 class GnuCCompiler(CCompiler):
-    std_warn_args = ['-Wall', '-Wpedantic', '-Winvalid-pch']
+    std_warn_args = []
+    old_warn = ['-Wall', '-pedantic', '-Winvalid-pch']
+    new_warn = ['-Wall', '-Wpedantic', '-Winvalid-pch']
 
     def __init__(self, exelist, version, gcc_type, is_cross, exe_wrapper=None):
         CCompiler.__init__(self, exelist, version, is_cross, exe_wrapper)
         self.id = 'gcc'
         self.gcc_type = gcc_type
+        if mesonlib.version_compare(version, ">=4.9.0"):
+            GnuCCompiler.std_warn_args= GnuCCompiler.new_warn
+        else:
+            GnuCCompiler.std_warn_args = GnuCCompiler.old_warn
 
     def get_always_args(self):
         return ['-pipe']
@@ -1032,7 +1051,7 @@ class GnuCCompiler(CCompiler):
         return get_gcc_soname_args(self.gcc_type, shlib_name, path, soversion)
 
     def can_compile(self, filename):
-        return super().can_compile(filename) or filename.split('.')[-1] == 's' # Gcc can do asm, too.
+        return super().can_compile(filename) or filename.split('.')[-1].lower() == 's' # Gcc can do asm, too.
 
 class GnuObjCCompiler(ObjCCompiler):
     std_warn_args = ['-Wall', '-Wpedantic', '-Winvalid-pch']
@@ -1109,7 +1128,7 @@ class ClangCCompiler(CCompiler):
         return 'pch'
 
     def can_compile(self, filename):
-        return super().can_compile(filename) or filename.split('.')[-1] == 's' # Clang can do asm, too.
+        return super().can_compile(filename) or filename.split('.')[-1].lower() == 's' # Clang can do asm, too.
 
     def get_pch_use_args(self, pch_dir, header):
         # Workaround for Clang bug http://llvm.org/bugs/show_bug.cgi?id=15136
@@ -1119,7 +1138,9 @@ class ClangCCompiler(CCompiler):
 
 
 class GnuCPPCompiler(CPPCompiler):
-    std_warn_args = ['-Wall', '-Wpedantic', '-Winvalid-pch', '-Wnon-virtual-dtor']
+    std_warn_args = []
+    new_warn = ['-Wall', '-Wpedantic', '-Winvalid-pch', '-Wnon-virtual-dtor']
+    old_warn = ['-Wall', '-pedantic', '-Winvalid-pch', '-Wnon-virtual-dtor']
     # may need to separate the latter to extra_debug_args or something
     std_debug_args = ['-g']
 
@@ -1127,6 +1148,10 @@ class GnuCPPCompiler(CPPCompiler):
         CPPCompiler.__init__(self, exelist, version, is_cross, exe_wrap)
         self.id = 'gcc'
         self.gcc_type = gcc_type
+        if mesonlib.version_compare(version, ">=4.9.0"):
+            GnuCPPCompiler.std_warn_args= GnuCPPCompiler.new_warn
+        else:
+            GnuCPPCompiler.std_warn_args = GnuCPPCompiler.old_warn
 
     def get_always_args(self):
         return ['-pipe']
@@ -1259,6 +1284,8 @@ end program prog
         return ['-o', outputname]
 
     def can_compile(self, src):
+        if hasattr(src, 'fname'):
+            src = src.fname
         suffix = os.path.splitext(src)[1].lower()
         if suffix == '.f' or suffix == '.f95' or suffix == '.f90':
             return True
