@@ -19,6 +19,8 @@
 # Currently one file, should probably be split into a
 # package before this gets too big.
 
+import re
+import platform
 import os, stat, glob, subprocess, shutil
 from coredata import MesonException
 import mlog
@@ -28,17 +30,10 @@ class DependencyException(MesonException):
     def __init__(self, *args, **kwargs):
         MesonException.__init__(self, *args, **kwargs)
 
-class CustomRule:
-    def __init__(self, cmd_list, name_templ, src_keyword, name, description):
-        self.cmd_list = cmd_list
-        self.name_templ = name_templ
-        self.src_keyword = src_keyword
-        self.name = name
-        self.description = description
-
 class Dependency():
     def __init__(self):
         self.name = "null"
+        self.is_found = False
 
     def get_compile_args(self):
         return []
@@ -47,7 +42,7 @@ class Dependency():
         return []
 
     def found(self):
-        return False
+        return self.is_found
 
     def get_sources(self):
         """Source files that need to be added to the target.
@@ -66,6 +61,9 @@ class Dependency():
 class PkgConfigDependency(Dependency):
     pkgconfig_found = None
 
+    __libtool_pat = re.compile("dlname='([A-z0-9\.\-\+]+)'\n")
+
+
     def __init__(self, name, kwargs):
         required = kwargs.get('required', True)
         Dependency.__init__(self)
@@ -84,7 +82,6 @@ class PkgConfigDependency(Dependency):
                                  stderr=subprocess.PIPE)
         out = p.communicate()[0]
         if p.returncode != 0:
-            mlog.log('Dependency', name, 'found:', mlog.red('NO'))
             if required:
                 raise DependencyException('Required dependency %s not found.' % name)
             self.modversion = 'none'
@@ -116,7 +113,21 @@ class PkgConfigDependency(Dependency):
             out = p.communicate()[0]
             if p.returncode != 0:
                 raise RuntimeError('Could not generate libs for %s.' % name)
-            self.libs = out.decode().split()
+            self.libs = []
+            for lib in out.decode().split():
+                if lib.endswith(".la"):
+                    shared_libname = self.__extract_libtool_shlib(lib)
+                    shared_lib = os.path.join(os.path.dirname(lib), shared_libname)
+                    if not os.path.exists(shared_lib):
+                        shared_lib = os.path.join(os.path.dirname(lib), ".libs", shared_libname)
+
+                    if not os.path.exists(shared_lib):
+                        raise RuntimeError('Got a libtools specific "%s" dependencies'
+                                           'but we could not compute the actual shared'
+                                           'library path' % lib)
+                    lib = shared_lib
+
+                self.libs.append(lib)
 
     def get_modversion(self):
         return self.modversion
@@ -145,6 +156,37 @@ class PkgConfigDependency(Dependency):
     def found(self):
         return self.is_found
 
+    def __extract_dlname_field(self, la_file):
+        f = open(la_file)
+        data = f.read()
+        f.close()
+        m = self.__libtool_pat.search(data)
+        if m:
+            return m.groups()[0]
+        else:
+            return None
+
+    def __extract_libtool_shlib(self, la_file):
+        '''
+        Returns the path to the shared library
+        corresponding to this .la file
+        '''
+        dlname = self.__extract_dlname_field(la_file)
+        if dlname is None:
+            return None
+
+        # Darwin uses absolute paths where possible; since the libtool files never
+        # contain absolute paths, use the libdir field
+        if platform.system() == 'Darwin':
+            dlbasename = os.path.basename(dlname)
+            libdir = self._extract_libdir_field(la_file)
+            if libdir is None:
+                return dlbasename
+            return libdir + '/' + dlbasename
+        # From the comments in extract_libtool(), older libtools had
+        # a path rather than the raw dlname
+        return os.path.basename(dlname)
+
 class WxDependency(Dependency):
     wx_found = None
 
@@ -173,6 +215,7 @@ class WxDependency(Dependency):
                     return
             mlog.log('Dependency wxwidgets found:', mlog.green('YES'))
             self.is_found = True
+            self.requested_modules = self.get_requested(kwargs)
             # wx-config seems to have a cflags as well but since it requires C++,
             # this should be good, at least for now.
             p = subprocess.Popen([self.wxc, '--cxxflags'], stdout=subprocess.PIPE,
@@ -182,13 +225,25 @@ class WxDependency(Dependency):
                 raise RuntimeError('Could not generate cargs for wxwidgets.')
             self.cargs = out.decode().split()
 
-            p = subprocess.Popen([self.wxc, '--libs'], stdout=subprocess.PIPE,
-                                 stderr=subprocess.PIPE)
+            p = subprocess.Popen([self.wxc, '--libs'] + self.requested_modules,
+                                 stdout=subprocess.PIPE, stderr=subprocess.PIPE)
             out = p.communicate()[0]
             if p.returncode != 0:
                 raise RuntimeError('Could not generate libs for wxwidgets.')
             self.libs = out.decode().split()
 
+    def get_requested(self, kwargs):
+        modules = 'modules'
+        if not modules in kwargs:
+            return []
+        candidates = kwargs[modules]
+        if isinstance(candidates, str):
+            return [candidates]
+        for c in candidates:
+            if not isinstance(c, str):
+                raise DependencyException('wxwidgets module argument is not a string.')
+        return candidates
+
     def get_modversion(self):
         return self.modversion
 
@@ -430,8 +485,8 @@ class GTestDependency(Dependency):
         self.include_dir = '/usr/include'
         self.src_include_dir = '/usr/src/gtest'
         self.src_dir = '/usr/src/gtest/src'
-        self.all_src = os.path.join(self.src_dir, 'gtest-all.cc')
-        self.main_src = os.path.join(self.src_dir, 'gtest_main.cc')
+        self.all_src = mesonlib.File.from_absolute_file(os.path.join(self.src_dir, 'gtest-all.cc'))
+        self.main_src = mesonlib.File.from_absolute_file(os.path.join(self.src_dir, 'gtest_main.cc'))
         self.detect()
 
     def found(self):
@@ -512,8 +567,8 @@ class GMockDependency(Dependency):
                 # versions of gmock that do different things.
                 self.compile_args = ['-I/usr/src/gmock', '-I/usr/src/gmock/src']
                 self.link_args = []
-                all_src = os.path.join(d, 'gmock-all.cc')
-                main_src = os.path.join(d, 'gmock_main.cc')
+                all_src = mesonlib.File.from_absolute_file(os.path.join(d, 'gmock-all.cc'))
+                main_src = mesonlib.File.from_absolute_file(os.path.join(d, 'gmock_main.cc'))
                 if kwargs.get('main', False):
                     self.sources = [all_src, main_src]
                 else:
@@ -544,14 +599,117 @@ class Qt5Dependency(Dependency):
         Dependency.__init__(self)
         self.name = 'qt5'
         self.root = '/usr'
+        mods = kwargs.get('modules', [])
+        self.cargs = []
+        self.largs= []
+        self.is_found = False
+        if isinstance(mods, str):
+            mods = [mods]
+        if len(mods) == 0:
+            raise DependencyException('No Qt5 modules specified.')
+        if shutil.which('pkg-config') is not None:
+            self.pkgconfig_detect(mods, kwargs)
+        elif shutil.which('qmake') is not None:
+            self.qmake_detect(mods, kwargs)
+        if not self.is_found:
+            mlog.log('Qt5 dependency found: ', mlog.red('NO'))
+        else:
+            mlog.log('Qt5 dependency found: ', mlog.green('YES'))
+
+    def pkgconfig_detect(self, mods, kwargs):
+        modules = []
+        for module in mods:
+            modules.append(PkgConfigDependency('Qt5' + module, kwargs))
+        for m in modules:
+            self.cargs += m.get_compile_args()
+            self.largs += m.get_link_args()
+        self.is_found = True
+
+    def qmake_detect(self, mods, kwargs):
+        pc = subprocess.Popen(['qmake', '-v'], stdout=subprocess.PIPE,
+                              stderr=subprocess.PIPE)
+        (stdo, _) = pc.communicate()
+        if pc.returncode != 0:
+            return
+        stdo = stdo.decode()
+        if not 'version 5' in stdo:
+            mlog.log('QMake is not for Qt5.')
+            return
+        (stdo, _) = subprocess.Popen(['qmake', '-query'], stdout=subprocess.PIPE).communicate()
+        qvars = {}
+        for line in stdo.decode().split('\n'):
+            line = line.strip()
+            if line == '':
+                continue
+            (k, v) = tuple(line.split(':', 1))
+            qvars[k] = v
+        if mesonlib.is_osx():
+            return self.framework_detect(qvars, mods, kwargs)
+        incdir = qvars['QT_INSTALL_HEADERS']
+        self.cargs.append('-I' + incdir)
+        libdir = qvars['QT_INSTALL_LIBS']
+        bindir = qvars['QT_INSTALL_BINS']
+        #self.largs.append('-L' + libdir)
+        for module in mods:
+            mincdir = os.path.join(incdir, 'Qt' + module)
+            self.cargs.append('-I' + mincdir)
+            libfile = os.path.join(libdir, 'Qt5' + module + '.lib')
+            if not os.path.isfile(libfile):
+                # MinGW links directly to .dll, not to .lib.
+                libfile = os.path.join(bindir, 'Qt5' + module + '.dll')
+            self.largs.append(libfile)
+        self.is_found = True
+
+    def framework_detect(self, qvars, modules, kwargs):
+        libdir = qvars['QT_INSTALL_LIBS']
+        for m in modules:
+            fname = 'Qt' + m
+            fwdep = ExtraFrameworkDependency(fname, kwargs.get('required', True), libdir)
+            self.cargs.append('-F' + libdir)
+            if fwdep.found():
+                self.is_found = True
+                self.cargs += fwdep.get_compile_args()
+                self.largs += fwdep.get_link_args()
+
+
+    def get_version(self):
+        return self.version
+
+    def get_compile_args(self):
+        return self.cargs
+
+    def get_sources(self):
+        return []
+
+    def get_link_args(self):
+        return self.largs
+
+    def found(self):
+        return self.is_found
+
+    def get_exe_args(self):
+        # Originally this was -fPIE but nowadays the default
+        # for upstream and distros seems to be -reduce-relocations
+        # which requires -fPIC. This may cause a performance
+        # penalty when using self-built Qt or on platforms
+        # where -fPIC is not required. If this is an issue
+        # for you, patches are welcome.
+        # Fix this to be more portable, especially to MSVC.
+        return ['-fPIC']
+
+class Qt4Dependency(Dependency):
+    def __init__(self, kwargs):
+        Dependency.__init__(self)
+        self.name = 'qt4'
+        self.root = '/usr'
         self.modules = []
         mods = kwargs.get('modules', [])
         if isinstance(mods, str):
             mods = [mods]
         for module in mods:
-            self.modules.append(PkgConfigDependency('Qt5' + module, kwargs))
+            self.modules.append(PkgConfigDependency('Qt' + module, kwargs))
         if len(self.modules) == 0:
-            raise DependencyException('No Qt5 modules specified.')
+            raise DependencyException('No Qt4 modules specified.')
 
     def get_version(self):
         return self.modules[0].get_version()
@@ -577,11 +735,6 @@ class Qt5Dependency(Dependency):
                 return False
         return True
 
-    def get_exe_args(self):
-        # Qt5 seems to require this always.
-        # Fix this to be more portable, especially to MSVC.
-        return ['-fPIE']
-
 class GnuStepDependency(Dependency):
     def __init__(self, kwargs):
         Dependency.__init__(self)
@@ -593,7 +746,6 @@ class GnuStepDependency(Dependency):
         try:
             gp = subprocess.Popen([confprog, '--help'],
                                   stdout=subprocess.PIPE, stderr=subprocess.PIPE)
-            
             gp.communicate()
         except FileNotFoundError:
             self.args = None
@@ -675,6 +827,31 @@ class AppleFrameworks(Dependency):
     def found(self):
         return mesonlib.is_osx()
 
+class GLDependency(Dependency):
+    def __init__(self, kwargs):
+        Dependency.__init__(self)
+        self.is_found = False
+        self.cargs = []
+        self.linkargs = []
+        try:
+            pcdep = PkgConfigDependency('gl', kwargs)
+            if pcdep.found():
+                self.is_found = True
+                self.cargs = pcdep.get_compile_args()
+                self.linkargs = pcdep.get_link_args()
+        except Exception:
+            pass
+        if mesonlib.is_osx():
+            self.is_found = True
+            self.linkargs = ['-framework', 'OpenGL']
+            return
+        if mesonlib.is_windows():
+            self.is_found = True
+            return
+
+    def get_link_args(self):
+        return self.linkargs
+
 # There are three different ways of depending on SDL2:
 # sdl2-config, pkg-config and OSX framework
 class SDL2Dependency(Dependency):
@@ -722,18 +899,21 @@ class SDL2Dependency(Dependency):
         return self.is_found
 
 class ExtraFrameworkDependency(Dependency):
-    def __init__(self, name, required):
+    def __init__(self, name, required, path=None):
         Dependency.__init__(self)
         self.name = None
-        self.detect(name)
+        self.detect(name, path)
         if self.found():
             mlog.log('Dependency', mlog.bold(name), 'found:', mlog.green('YES'), os.path.join(self.path, self.name))
         else:
             mlog.log('Dependency', name, 'found:', mlog.red('NO'))
 
-    def detect(self, name):
+    def detect(self, name, path):
         lname = name.lower()
-        paths = ['/Library/Frameworks']
+        if path is None:
+            paths = ['/Library/Frameworks']
+        else:
+            paths = [path]
         for p in paths:
             for d in os.listdir(p):
                 fullpath = os.path.join(p, d)
@@ -746,10 +926,14 @@ class ExtraFrameworkDependency(Dependency):
                 return
 
     def get_compile_args(self):
-        return ['-I' + os.path.join(self.path, self.name, 'Headers')]
+        if self.found():
+            return ['-I' + os.path.join(self.path, self.name, 'Headers')]
+        return []
 
     def get_link_args(self):
-        return ['-F' + self.path, '-framework', self.name.split('.')[0]]
+        if self.found():
+            return ['-F' + self.path, '-framework', self.name.split('.')[0]]
+        return []
 
     def found(self):
         return self.name is not None
@@ -782,9 +966,13 @@ def find_external_dependency(name, kwargs):
     except Exception as e:
         pkg_exc = e
     if mesonlib.is_osx():
-        return ExtraFrameworkDependency(name, required)
+        fwdep = ExtraFrameworkDependency(name, required)
+        if required and not fwdep.found():
+            raise DependencyException('Dependency "%s" not found' % name)
+        return fwdep
     if pkg_exc is not None:
         raise pkg_exc
+    mlog.log('Dependency', mlog.bold(name), 'found:', mlog.red('NO'))
     return pkgdep
 
 # This has to be at the end so the classes it references
@@ -793,8 +981,10 @@ packages = {'boost': BoostDependency,
             'gtest': GTestDependency,
             'gmock': GMockDependency,
             'qt5': Qt5Dependency,
+            'qt4': Qt4Dependency,
             'gnustep': GnuStepDependency,
             'appleframeworks': AppleFrameworks,
             'wxwidgets' : WxDependency,
             'sdl2' : SDL2Dependency,
+            'gl' : GLDependency,
             }
