@@ -77,7 +77,7 @@ class Environment():
 
         # List of potential compilers.
         if mesonlib.is_windows():
-            self.default_c = ['cl', 'cc']
+            self.default_c = ['cl', 'cc', 'gcc']
             self.default_cpp = ['cl', 'c++']
         else:
             self.default_c = ['cc']
@@ -547,8 +547,9 @@ class Environment():
     def get_datadir(self):
         return self.coredata.datadir
 
-    def find_library(self, libname):
-        dirs = mesonlib.get_library_dirs()
+    def find_library(self, libname, dirs):
+        if dirs is None:
+            dirs = mesonlib.get_library_dirs()
         suffixes = [self.get_shared_lib_suffix(), self.get_static_lib_suffix()]
         prefix = self.get_shared_lib_prefix()
         for d in dirs:
