@@ -22,6 +22,9 @@ import optinterpreter
 import wrap
 import mesonlib
 import os, sys, platform, subprocess, shutil, uuid, re
+from mesonlib import File
+from functools import wraps
+
 import importlib
 
 class InterpreterException(coredata.MesonException):
@@ -33,6 +36,39 @@ class InvalidCode(InterpreterException):
 class InvalidArguments(InterpreterException):
     pass
 
+# Decorators for method calls.
+
+def check_stringlist(a, msg='Arguments must be strings.'):
+    if not isinstance(a, list):
+        raise InvalidArguments('Argument not a list.')
+    if not all(isinstance(s, str) for s in a):
+        raise InvalidArguments(msg)
+
+def noKwargs(f):
+    @wraps(f)
+    def wrapped(self, node, args, kwargs):
+        if len(kwargs) != 0:
+            raise InvalidArguments('Function does not take keyword arguments.')
+        return f(self, node, args, kwargs)
+    return wrapped
+
+def stringArgs(f):
+    @wraps(f)
+    def wrapped(self, node, args, kwargs):
+        assert(isinstance(args, list))
+        check_stringlist(args)
+        return f(self, node, args, kwargs)
+    return wrapped
+
+def stringifyUserArguments(args):
+    if isinstance(args, list):
+        return '[%s]' % ', '.join([stringifyUserArguments(x) for x in args])
+    elif isinstance(args, int):
+        return str(args)
+    elif isinstance(args, str):
+        return "'%s'" % args
+    raise InvalidArguments('Function accepts only strings, integers, lists and lists thereof.')
+
 class InterpreterObject():
     def __init__(self):
         self.methods = {}
@@ -234,13 +270,7 @@ class GeneratorHolder(InterpreterObject):
     def process_method(self, args, kwargs):
         if len(kwargs) > 0:
             raise InvalidArguments('Process does not take keyword arguments.')
-        if isinstance(args, str):
-            args = [args]
-        if not isinstance(args, list):
-            raise InvalidArguments('Argument to "process" must be a string or a list of strings.')
-        for a in args:
-            if not isinstance(a, str):
-                raise InvalidArguments('A non-string object in "process" arguments.')
+        check_stringlist(args)
         gl = GeneratedListHolder(self)
         [gl.add_file(os.path.join(self.interpreter.subdir, a)) for a in args]
         return gl
@@ -284,9 +314,9 @@ class Host(InterpreterObject):
         return sys.byteorder != 'little'
 
 class IncludeDirsHolder(InterpreterObject):
-    def __init__(self, curdir, dirs, kwargs):
+    def __init__(self, curdir, dirs):
         super().__init__()
-        self.held_object = build.IncludeDirs(curdir, dirs, kwargs)
+        self.held_object = build.IncludeDirs(curdir, dirs)
 
 class Headers(InterpreterObject):
 
@@ -325,9 +355,7 @@ class Data(InterpreterObject):
         if not isinstance(kwsource, list):
             kwsource = [kwsource]
         self.sources += kwsource
-        for s in self.sources:
-            if not isinstance(s, str):
-                raise InterpreterException('Install file name must be a string.')
+        check_stringlist(self.sources)
         self.install_dir = kwargs.get('install_dir', None)
         if not isinstance(self.install_dir, str):
             raise InterpreterException('Custom_install_dir must be a string.')
@@ -474,19 +502,23 @@ class CompilerHolder(InterpreterObject):
                              'run' : self.run_method,
                              'has_function' : self.has_function_method,
                              'has_member' : self.has_member_method,
+                             'has_type' : self.has_type_method,
                              'alignment' : self.alignment_method,
-                             'version' : self.version_method
+                             'version' : self.version_method,
+                             'cmd_array' : self.cmd_array_method,
                              })
 
     def version_method(self, args, kwargs):
         return self.compiler.version
 
+    def cmd_array_method(self, args, kwargs):
+        return self.compiler.exelist
+
     def alignment_method(self, args, kwargs):
         if len(args) != 1:
             raise InterpreterException('Alignment method takes exactly one positional argument.')
+        check_stringlist(args)
         typename = args[0]
-        if not isinstance(typename, str):
-            raise InterpreterException('First argument is not a string.')
         result = self.compiler.alignment(typename, self.environment)
         mlog.log('Checking for alignment of "', mlog.bold(typename), '": ', result, sep='')
         return result
@@ -494,12 +526,11 @@ class CompilerHolder(InterpreterObject):
     def run_method(self, args, kwargs):
         if len(args) != 1:
             raise InterpreterException('Run method takes exactly one positional argument.')
+        check_stringlist(args)
         code = args[0]
         testname = kwargs.get('name', '')
         if not isinstance(testname, str):
             raise InterpreterException('Testname argument must be a string.')
-        if not isinstance(code, str):
-            raise InterpreterException('First argument is not a string.')
         result = self.compiler.run(code)
         if len(testname) > 0:
             if not result.compiled:
@@ -517,12 +548,9 @@ class CompilerHolder(InterpreterObject):
     def has_member_method(self, args, kwargs):
         if len(args) != 2:
             raise InterpreterException('Has_member takes exactly two arguments.')
+        check_stringlist(args)
         typename = args[0]
-        if not isinstance(typename, str):
-            raise InterpreterException('Name of type must be a string.')
         membername = args[1]
-        if not isinstance(membername, str):
-            raise InterpreterException('Name of member must be a string.')
         prefix = kwargs.get('prefix', '')
         if not isinstance(prefix, str):
             raise InterpreterException('Prefix argument of has_function must be a string.')
@@ -538,9 +566,8 @@ class CompilerHolder(InterpreterObject):
     def has_function_method(self, args, kwargs):
         if len(args) != 1:
             raise InterpreterException('Has_function takes exactly one argument.')
+        check_stringlist(args)
         funcname = args[0]
-        if not isinstance(funcname, str):
-            raise InterpreterException('Argument to has_function must be a string.')
         prefix = kwargs.get('prefix', '')
         if not isinstance(prefix, str):
             raise InterpreterException('Prefix argument of has_function must be a string.')
@@ -552,12 +579,27 @@ class CompilerHolder(InterpreterObject):
         mlog.log('Checking for function "', mlog.bold(funcname), '": ', hadtxt, sep='')
         return had
 
+    def has_type_method(self, args, kwargs):
+        if len(args) != 1:
+            raise InterpreterException('Has_type takes exactly one argument.')
+        check_stringlist(args)
+        typename = args[0]
+        prefix = kwargs.get('prefix', '')
+        if not isinstance(prefix, str):
+            raise InterpreterException('Prefix argument of has_type must be a string.')
+        had = self.compiler.has_type(typename, prefix)
+        if had:
+            hadtxt = mlog.green('YES')
+        else:
+            hadtxt = mlog.red('NO')
+        mlog.log('Checking for type "', mlog.bold(typename), '": ', hadtxt, sep='')
+        return had
+
     def sizeof_method(self, args, kwargs):
         if len(args) != 1:
             raise InterpreterException('Sizeof takes exactly one argument.')
+        check_stringlist(args)
         element = args[0]
-        if not isinstance(element, str):
-            raise InterpreterException('Argument to sizeof must be a string.')
         prefix = kwargs.get('prefix', '')
         if not isinstance(prefix, str):
             raise InterpreterException('Prefix argument of sizeof must be a string.')
@@ -568,12 +610,11 @@ class CompilerHolder(InterpreterObject):
     def compiles_method(self, args, kwargs):
         if len(args) != 1:
             raise InterpreterException('compiles method takes exactly one argument.')
+        check_stringlist(args)
         string = args[0]
         testname = kwargs.get('name', '')
         if not isinstance(testname, str):
             raise InterpreterException('Testname argument must be a string.')
-        if not isinstance(string, str):
-            raise InterpreterException('Argument to compiles() must be a string')
         result = self.compiler.compiles(string)
         if len(testname) > 0:
             if result:
@@ -586,9 +627,8 @@ class CompilerHolder(InterpreterObject):
     def has_header_method(self, args, kwargs):
         if len(args) != 1:
             raise InterpreterException('has_header method takes exactly one argument.')
+        check_stringlist(args)
         string = args[0]
-        if not isinstance(string, str):
-            raise InterpreterException('Argument to has_header() must be a string')
         haz = self.compiler.has_header(string)
         if haz:
             h = mlog.green('YES')
@@ -645,9 +685,8 @@ class MesonMain(InterpreterObject):
     def set_install_script_method(self, args, kwargs):
         if len(args) != 1:
             raise InterpreterException('Set_install_script takes exactly one argument.')
+        check_stringlist(args)
         scriptbase = args[0]
-        if not isinstance(scriptbase, str):
-            raise InterpreterException('Set_install_script argument is not a string.')
         scriptfile = os.path.join(self.interpreter.environment.source_dir,
                                   self.interpreter.subdir, scriptbase)
         if not os.path.isfile(scriptfile):
@@ -701,15 +740,16 @@ class MesonMain(InterpreterObject):
         return self.build.environment.coredata.unity
 
     def is_subproject_method(self, args, kwargs):
-        return self.interpreter.subproject != ''
+        return self.interpreter.is_subproject()
 
 class Interpreter():
 
-    def __init__(self, build, subproject='', subdir=''):
+    def __init__(self, build, subproject='', subdir='', subproject_dir='subprojects'):
         self.build = build
         self.subproject = subproject
         self.subdir = subdir
         self.source_root = build.environment.get_source_dir()
+        self.subproject_dir = subproject_dir
         option_file = os.path.join(self.source_root, self.subdir, 'meson_options.txt')
         if os.path.exists(option_file):
             oi = optinterpreter.OptionInterpreter(self.subproject,\
@@ -777,6 +817,7 @@ class Interpreter():
                       'vcs_tag' : self.func_vcs_tag,
                       'set_variable' : self.func_set_variable,
                       'import' : self.func_import,
+                      'files' : self.func_files,
                       }
 
     def module_method_callback(self, invalues):
@@ -866,17 +907,22 @@ class Interpreter():
         value = self.to_native(args[1])
         self.set_variable(varname, value)
 
+    @stringArgs
+    @noKwargs
     def func_import(self, node, args, kwargs):
         if len(args) != 1:
             raise InvalidCode('Import takes one argument.')
         modname = args[0]
-        if not isinstance(modname, str):
-            raise InvalidCode('Argument to import was not a string')
         if not modname in self.environment.coredata.modules:
             module = importlib.import_module('modules.' + modname).initialize()
             self.environment.coredata.modules[modname] = module
         return ModuleHolder(modname, self.environment.coredata.modules[modname], self)
 
+    @stringArgs
+    @noKwargs
+    def func_files(self, node, args, kwargs):
+        return [File.from_source_file(self.environment.source_dir, self.subdir, fname) for fname in args]
+
     def set_variable(self, varname, variable):
         if variable is None:
             raise InvalidCode('Can not assign None to variable.')
@@ -917,6 +963,8 @@ class Interpreter():
             return self.evaluate_orstatement(cur)
         elif isinstance(cur, mparser.NotNode):
             return self.evaluate_notstatement(cur)
+        elif isinstance(cur, mparser.UMinusNode):
+            return self.evaluate_uminusstatement(cur)
         elif isinstance(cur, mparser.ArithmeticNode):
             return self.evaluate_arithmeticstatement(cur)
         elif isinstance(cur, mparser.ForeachClauseNode):
@@ -945,13 +993,11 @@ class Interpreter():
         cargs = args[1:]
         if isinstance(cmd, ExternalProgramHolder):
             cmd = cmd.get_command()
-        elif not isinstance(cmd, str):
-            raise InterpreterException('First argument is of incorrect type.')
-        for i in cargs:
-            if not isinstance(i, str):
-                raise InterpreterException('Run_command arguments must be strings.')
-        if not isinstance(cmd, list):
+        elif isinstance(cmd, str):
             cmd = [cmd]
+        else:
+            raise InterpreterException('First argument is of incorrect type.')
+        check_stringlist(cargs, 'Run_command arguments must be strings.')
         args = cmd + cargs
         in_builddir = kwargs.get('in_builddir', False)
         if not isinstance(in_builddir, bool):
@@ -959,15 +1005,13 @@ class Interpreter():
         return RunProcess(args, self.environment.source_dir, self.environment.build_dir,
                           self.subdir, in_builddir)
 
+    @stringArgs
     def func_gettext(self, nodes, args, kwargs):
         if len(args) != 1:
             raise InterpreterException('Gettext requires one positional argument (package name).')
         packagename = args[0]
-        if not isinstance(packagename, str):
-            raise InterpreterException('Gettext argument is not a string.')
         languages = kwargs.get('languages', None)
-        if not isinstance(languages, list):
-            raise InterpreterException('Argument languages must be a list of strings.')
+        check_stringlist(languages, 'Argument languages must be a list of strings.')
         # TODO: check that elements are strings
         if len(self.build.pot) > 0:
             raise InterpreterException('More than one gettext definition currently not supported.')
@@ -1006,48 +1050,50 @@ class Interpreter():
         p = build.PkgConfigGenerator(libs, subdirs, name, description, version, filebase)
         self.build.pkgconfig_gens.append(p)
 
+    @stringArgs
+    @noKwargs
     def func_subproject(self, nodes, args, kwargs):
         if len(args) != 1:
             raise InterpreterException('Subproject takes exactly one argument')
         dirname = args[0]
-        if not isinstance(dirname, str):
-            raise InterpreterException('Subproject argument must be a string')
         if self.subdir != '':
             segs = os.path.split(self.subdir)
-            if len(segs) != 2 or segs[0] != 'subprojects':
+            if len(segs) != 2 or segs[0] != self.subproject_dir:
                 raise InterpreterException('Subprojects must be defined at the root directory.')
         if dirname in self.subproject_stack:
             fullstack = self.subproject_stack + [dirname]
             incpath = ' => '.join(fullstack)
             raise InterpreterException('Recursive include of subprojects: %s.' % incpath)
-        if dirname == self.subprojects:
+        if dirname in self.subprojects:
             return self.subprojects[dirname]
-        subdir = os.path.join('subprojects', dirname)
-        r = wrap.Resolver(os.path.join(self.build.environment.get_source_dir(), 'subprojects'))
+        subdir = os.path.join(self.subproject_dir, dirname)
+        r = wrap.Resolver(os.path.join(self.build.environment.get_source_dir(), self.subproject_dir))
         resolved = r.resolve(dirname)
         if resolved is None:
             raise InterpreterException('Subproject directory does not exist and can not be downloaded.')
-        subdir = os.path.join('subprojects', resolved)
+        subdir = os.path.join(self.subproject_dir, resolved)
         os.makedirs(os.path.join(self.build.environment.get_build_dir(), subdir), exist_ok=True)
         self.global_args_frozen = True
         mlog.log('\nExecuting subproject ', mlog.bold(dirname), '.\n', sep='')
-        subi = Interpreter(self.build, dirname, subdir)
+        subi = Interpreter(self.build, dirname, subdir, self.subproject_dir)
+        subi.subprojects = self.subprojects
 
         subi.subproject_stack = self.subproject_stack + [dirname]
         subi.run()
         mlog.log('\nSubproject', mlog.bold(dirname), 'finished.')
         self.build.subprojects[dirname] = True
+        self.subprojects.update(subi.subprojects)
         self.subprojects[dirname] = SubprojectHolder(subi)
         self.build_def_files += subi.build_def_files
         return self.subprojects[dirname]
 
+    @stringArgs
+    @noKwargs
     def func_get_option(self, nodes, args, kwargs):
         if len(args) != 1:
             raise InterpreterException('Argument required for get_option.')
         optname = args[0]
-        if not isinstance(optname, str):
-            raise InterpreterException('Argument of get_option must be a string.')
-        if self.subproject != '':
+        if self.is_subproject():
             optname = self.subproject + ':' + optname
         try:
             return self.environment.get_coredata().get_builtin_option(optname)
@@ -1057,23 +1103,26 @@ class Interpreter():
             raise InterpreterException('Tried to access unknown option "%s".' % optname)
         return self.environment.coredata.user_options[optname].value
 
+    @noKwargs
     def func_configuration_data(self, node, args, kwargs):
         if len(args) != 0:
             raise InterpreterException('configuration_data takes no arguments')
         return ConfigurationDataHolder()
 
+    @stringArgs
     def func_project(self, node, args, kwargs):
         if len(args)< 2:
             raise InvalidArguments('Not enough arguments to project(). Needs at least the project name and one language')
-        if len(kwargs) > 0:
-            raise InvalidArguments('Project() does not take keyword arguments.')
-        for a in args:
-            if not isinstance(a, str):
-                raise InvalidArguments('Argument %s is not a string.' % str(a))
-        if self.subproject == '':
+        if list(kwargs.keys()) != ['subproject_dir'] and len(kwargs) != 0:
+            raise InvalidArguments('project() only accepts the keyword argument "subproject_dir"')
+
+        if not self.is_subproject():
             self.build.project_name = args[0]
         if self.subproject in self.build.projects:
             raise InvalidCode('Second call to project().')
+        if not self.is_subproject() and 'subproject_dir' in kwargs:
+            self.subproject_dir = kwargs['subproject_dir']
+
         self.build.projects[self.subproject] = args[0]
         mlog.log('Project name: ', mlog.bold(args[0]), sep='')
         self.add_languages(node, args[1:])
@@ -1082,10 +1131,28 @@ class Interpreter():
             if not 'c' in langs:
                 raise InterpreterException('Compiling Vala requires a C compiler')
 
+    @noKwargs
     def func_message(self, node, args, kwargs):
-        self.validate_arguments(args, 1, [str])
-        mlog.log(mlog.bold('Message:'), args[0])
+        # reduce arguments again to avoid flattening posargs
+        (posargs, kwargs) = self.reduce_arguments(node.args)
+        if len(posargs) != 1:
+            raise InvalidArguments('Expected 1 argument, got %d' % len(posargs))
 
+        arg = posargs[0]
+        if isinstance(arg, list):
+            argstr =  stringifyUserArguments(arg)
+        elif isinstance(arg, str):
+            argstr = arg
+        elif isinstance(arg, int):
+            argstr = str(arg)
+        else:
+            raise InvalidArguments('Function accepts only strings, integers, lists and lists thereof.')
+
+        mlog.log(mlog.bold('Message:'), argstr)
+        return
+
+
+    @noKwargs
     def func_error(self, node, args, kwargs):
         self.validate_arguments(args, 1, [str])
         raise InterpreterException('Error encountered: ' + args[0])
@@ -1178,7 +1245,18 @@ class Interpreter():
         if libname in self.coredata.ext_libs and\
            self.coredata.ext_libs[libname].found():
             return ExternalLibraryHolder(self.coredata.ext_libs[libname])
-        result = self.environment.find_library(libname)
+        if 'dirs' in kwargs:
+            search_dirs = kwargs['dirs']
+            if not isinstance(search_dirs, list):
+                search_dirs = [search_dirs]
+            for i in search_dirs:
+                if not isinstance(i, str):
+                    raise InvalidCode('Directory entry is not a string.')
+                if not os.path.isabs(i):
+                    raise InvalidCode('Search directory %s is not an absolute path.' % i)
+        else:
+            search_dirs = None
+        result = self.environment.find_library(libname, search_dirs)
         extlib = dependencies.ExternalLibrary(libname, result)
         libobj = ExternalLibraryHolder(extlib)
         self.coredata.ext_libs[libname] = extlib
@@ -1238,12 +1316,11 @@ class Interpreter():
         kwargs.setdefault('build_always', True)
         return self.func_custom_target(node, [kwargs['output']], kwargs)
 
+    @stringArgs
     def func_custom_target(self, node, args, kwargs):
         if len(args) != 1:
             raise InterpreterException('Incorrect number of arguments')
         name = args[0]
-        if not isinstance(name, str):
-            raise InterpreterException('Argument must be a string.')
         if name in coredata.forbidden_target_names:
             raise InvalidArguments('Target name "%s" is reserved for Meson\'s internal use. Please rename.'\
                                    % name)
@@ -1253,12 +1330,11 @@ class Interpreter():
         self.build.targets[name] = tg.held_object
         return tg
 
+    @stringArgs
+    @noKwargs
     def func_run_target(self, node, args, kwargs):
         if len(args) < 2:
             raise InterpreterException('Incorrect number of arguments')
-        for i in args:
-            if not isinstance(i, str):
-                raise InvalidArguments('Argument is not a string')
         name = args[0]
         if name in coredata.forbidden_target_names:
             raise InvalidArguments('Target name "%s" is reserved for Meson\'s internal use. Please rename.'\
@@ -1318,30 +1394,24 @@ class Interpreter():
         self.build.tests.append(t)
         mlog.debug('Adding test "', mlog.bold(args[0]), '".', sep='')
 
+    @stringArgs
     def func_install_headers(self, node, args, kwargs):
-        args = self.flatten(args)
-        for a in args:
-            if not isinstance(a, str):
-                raise InvalidArguments('Argument %s is not a string.' % str(a))
         h = Headers(self.subdir, args, kwargs)
         self.build.headers.append(h)
         return h
 
+    @stringArgs
     def func_install_man(self, node, args, kwargs):
-        for a in args:
-            if not isinstance(a, str):
-                raise InvalidArguments('Argument %s is not a string.' % str(a))
         m = Man(self.subdir, args, kwargs)
         self.build.man.append(m)
         return m
 
+    @noKwargs
     def func_subdir(self, node, args, kwargs):
-        if len(kwargs) > 0:
-            raise InvalidArguments('subdir command takes no keyword arguments.')
         self.validate_arguments(args, 1, [str])
         if '..' in args[0]:
             raise InvalidArguments('Subdir contains ..')
-        if self.subdir == '' and args[0] == 'subprojects':
+        if self.subdir == '' and args[0] == self.subproject_dir:
             raise InvalidArguments('Must not go into subprojects dir with subdir(), use subproject() instead.')
         prev_subdir = self.subdir
         subdir = os.path.join(prev_subdir, args[0])
@@ -1369,20 +1439,16 @@ class Interpreter():
         self.evaluate_codeblock(codeblock)
         self.subdir = prev_subdir
 
+    @stringArgs
     def func_install_data(self, node, args, kwargs):
-        for a in args:
-            if not isinstance(a, str):
-                raise InvalidArguments('Argument %s is not a string.' % str(a))
         data = Data(True, self.subdir, args, kwargs)
         self.build.data.append(data)
         return data
 
+    @stringArgs
     def func_install_subdir(self, node, args, kwargs):
         if len(args ) != 1:
             raise InvalidArguments('Install_subdir requires exactly one argument.')
-        for a in args:
-            if not isinstance(a, str):
-                raise InvalidArguments('Argument %s is not a string.' % str(a))
         if not 'install_dir' in kwargs:
             raise InvalidArguments('Missing keyword argument install_dir')
         install_dir = kwargs['install_dir']
@@ -1433,21 +1499,19 @@ class Interpreter():
         outputfile = os.path.join(self.environment.build_dir, self.subdir, output)
         return outputfile
 
+    @stringArgs
+    @noKwargs
     def func_include_directories(self, node, args, kwargs):
         absbase = os.path.join(self.environment.get_source_dir(), self.subdir)
         for a in args:
-            if not isinstance(a, str):
-                raise InvalidArguments('Argument %s is not a string.' % str(a))
             absdir = os.path.join(absbase, a)
             if not os.path.isdir(absdir):
                 raise InvalidArguments('Include dir %s does not exist.' % a)
-        i = IncludeDirsHolder(self.subdir, args, kwargs)
+        i = IncludeDirsHolder(self.subdir, args)
         return i
 
+    @stringArgs
     def func_add_global_arguments(self, node, args, kwargs):
-        for a in args:
-            if not isinstance(a, str):
-                raise InvalidArguments('Argument %s is not a string.' % str(a))
         if self.subproject != '':
             raise InvalidCode('Global arguments can not be set in subprojects because there is no way to make that reliable.')
         if self.global_args_frozen:
@@ -1480,8 +1544,20 @@ class Interpreter():
                 result.append(a)
         return result
 
+    def source_strings_to_files(self, sources):
+        results = []
+        for s in sources:
+            if isinstance(s, File) or isinstance(s, GeneratedListHolder) or \
+            isinstance(s, CustomTargetHolder):
+                pass
+            elif isinstance(s, str):
+                s = File.from_source_file(self.environment.source_dir, self.subdir, s)
+            else:
+                raise RuntimeError("Unreachable code")
+            results.append(s)
+        return results
+
     def build_target(self, node, args, kwargs, targetholder):
-        args = self.flatten(args)
         name = args[0]
         sources = args[1:]
         if self.environment.is_cross_build():
@@ -1501,6 +1577,7 @@ class Interpreter():
         except KeyError:
             kw_src = []
         sources += kw_src
+        sources = self.source_strings_to_files(sources)
         objs = self.flatten(kwargs.get('objects', []))
         if not isinstance(objs, list):
             objs = [objs]
@@ -1538,7 +1615,7 @@ class Interpreter():
         func_name = node.func_name
         (posargs, kwargs) = self.reduce_arguments(node.args)
         if func_name in self.funcs:
-            return self.funcs[func_name](node, posargs, kwargs)
+            return self.funcs[func_name](node, self.flatten(posargs), kwargs)
         else:
             raise InvalidCode('Unknown function "%s".' % func_name)
 
@@ -1547,7 +1624,8 @@ class Interpreter():
             isinstance(value, dependencies.Dependency) or\
             isinstance(value, str) or\
             isinstance(value, int) or \
-            isinstance(value, list):
+            isinstance(value, list) or \
+            isinstance(value, File):
             return True
         return False
 
@@ -1574,6 +1652,8 @@ class Interpreter():
                 raise InvalidArguments('Keyword argument name is not a string.')
             a = args.kwargs[key]
             reduced_kw[key] = self.evaluate_statement(a)
+        if not isinstance(reduced_pos, list):
+            reduced_pos = [reduced_pos]
         return (reduced_pos, reduced_kw)
 
     def string_method_call(self, obj, method_name, args):
@@ -1642,7 +1722,7 @@ class Interpreter():
             index = args[0]
             if not isinstance(index, int):
                 raise InvalidArguments('Array index must be a number.')
-            if index < 0 or index >= len(obj):
+            if index < -len(obj) or index >= len(obj):
                 raise InvalidArguments('Array index %s is out of bounds for array of size %d.' % (index, len(obj)))
             return obj[index]
         raise InterpreterException('Arrays do not have a method called "%s".' % method_name)
@@ -1747,6 +1827,14 @@ class Interpreter():
             raise InterpreterException('Argument to "not" is not a boolean.')
         return not v
 
+    def evaluate_uminusstatement(self, cur):
+        v = self.evaluate_statement(cur.value)
+        if isinstance(v, mparser.NumberNode):
+            v = v.value
+        if not isinstance(v, int):
+            raise InterpreterException('Argument to negation is not an integer.')
+        return -v
+
     def evaluate_arithmeticstatement(self, cur):
         l = self.to_native(self.evaluate_statement(cur.left))
         r = self.to_native(self.evaluate_statement(cur.right))
@@ -1776,3 +1864,6 @@ class Interpreter():
         if len(kwargs) > 0:
             raise InvalidCode('Keyword arguments are invalid in array construction.')
         return arguments
+
+    def is_subproject(self):
+        return self.subproject != ''
