@@ -17,6 +17,7 @@ import environment, mesonlib
 import build
 import mlog
 import dependencies
+from mesonlib import File
 from meson_install import InstallData
 from build import InvalidArguments
 from coredata import MesonException
@@ -283,6 +284,12 @@ class NinjaBackend(backends.Backend):
         if target.build_always:
             deps.append('PHONY')
         elem = NinjaBuildElement(ofilenames, 'CUSTOM_COMMAND', deps)
+        for d in target.extra_depends:
+            tmp = d.get_filename()
+            if not isinstance(tmp, list):
+                tmp = [tmp]
+            for fname in tmp:
+                elem.add_dep(os.path.join(d.get_subdir(), fname))
         cmd = []
         for i in target.command:
             for (j, src) in enumerate(srcs):
@@ -603,7 +610,7 @@ class NinjaBackend(backends.Backend):
         src_list = target.get_sources()
         compiler = self.get_compiler_for_source(src_list[0])
         assert(compiler.get_language() == 'cs')
-        rel_srcs = [os.path.join(self.build_to_src, s) for s in src_list]
+        rel_srcs = [s.rel_to_builddir(self.build_to_src) for s in src_list]
         deps = []
         commands = target.extra_args.get('cs', [])
         commands += compiler.get_buildtype_args(buildtype)
@@ -630,12 +637,11 @@ class NinjaBackend(backends.Backend):
         elem.write(outfile)
 
     def generate_single_java_compile(self, subdir, src, target, compiler, outfile):
-        buildtype = self.environment.coredata.buildtype
         args = []
         args += compiler.get_buildtype_args(self.environment.coredata.buildtype)
         args += compiler.get_output_args(self.get_target_private_dir(target))
-        rel_src = os.path.join(self.build_to_src, subdir, src)
-        plain_class_path = src[:-4] + 'class'
+        rel_src = src.rel_to_builddir(self.build_to_src)
+        plain_class_path = src.fname[:-4] + 'class'
         rel_obj = os.path.join(self.get_target_private_dir(target), plain_class_path)
         element = NinjaBuildElement(rel_obj,
                     compiler.get_language() + '_COMPILER', rel_src)
@@ -657,27 +663,39 @@ class NinjaBackend(backends.Backend):
         for s in target.get_sources():
             if not s.endswith('.vala'):
                 continue
-            vapibase = os.path.basename(s)[:-4] + 'vapi'
+            vapibase = os.path.basename(s.fname)[:-4] + 'vapi'
             rel_vapi = os.path.join(self.get_target_dir(target), target.get_basename() + '.dir', vapibase)
             args = ['--fast-vapi=' + rel_vapi]
-            rel_s = os.path.join(self.build_to_src, s)
+            rel_s = s.rel_to_builddir(self.build_to_src)
             element = NinjaBuildElement(rel_vapi, valac.get_language() + '_COMPILER', rel_s)
             element.add_item('ARGS', args)
             element.write(outfile)
             fastvapis[s] = (vapibase, rel_vapi)
         return fastvapis
 
+    def split_vala_sources(self, sources):
+        src = []
+        vapi_src = []
+        for s in sources:
+            if s.endswith('.vapi'):
+                vapi_src.append(s)
+            else:
+                src.append(s)
+        return (src, vapi_src)
+
     def generate_vala_compile(self, target, outfile):
         """Vala is compiled into C. Set up all necessary build steps here."""
         valac = self.environment.coredata.compilers['vala']
         fast_vapis = self.generate_fastvapi_compile(target, valac, outfile)
         generated_c = []
-        for s in target.get_sources():
+        (src, vapi_src) = self.split_vala_sources(target.get_sources())
+        vapi_src = [x.rel_to_builddir(self.build_to_src) for x in vapi_src]
+        for s in src:
             if not s.endswith('.vala'):
                 continue
             args = ['-d', self.get_target_private_dir(target)]
-            sc = os.path.basename(s)[:-4] + 'c'
-            args += ['-C', '-o', sc]
+            sc = os.path.basename(s.fname)[:-4] + 'c'
+            args += ['-C']
             vapi_order_deps = []
             for (sourcefile, vapi_info) in fast_vapis.items():
                 if sourcefile == s:
@@ -686,13 +704,14 @@ class NinjaBackend(backends.Backend):
                 args += ['--use-fast-vapi=' + rel_vapi]
                 vapi_order_deps.append(rel_vapi)
             relsc = os.path.join(self.get_target_dir(target), target.get_basename() + '.dir', sc)
-            rel_s = os.path.join(self.build_to_src, s)
+            rel_s = s.rel_to_builddir(self.build_to_src)
             args += ['--deps', relsc + '.d']
             if self.environment.coredata.werror:
                 args += valac.get_werror_args()
             for d in target.external_deps:
                 if isinstance(d, dependencies.PkgConfigDependency):
                     args += ['--pkg', d.name]
+            args += vapi_src
             generated_c += [relsc]
             element = NinjaBuildElement(relsc, valac.get_language() + '_COMPILER', rel_s)
             element.add_item('ARGS', args)
@@ -771,9 +790,8 @@ class NinjaBackend(backends.Backend):
                 if is_cross:
                     crstr = '_CROSS'
                 rule = 'rule %s%s_LINKER\n' % (langname, crstr)
-                command = ' command = %s %s $ARGS  %s $in $LINK_ARGS $aliasing\n' % \
-                (execute_wrapper,
-                 ' '.join(compiler.get_linker_exelist()),\
+                command = ' command = %s $ARGS  %s $in $LINK_ARGS $aliasing\n' % \
+                (' '.join(compiler.get_linker_exelist()),\
                  ' '.join(compiler.get_linker_output_args('$out')))
                 description = ' description = Linking target $out'
                 outfile.write(rule)
@@ -992,6 +1010,7 @@ rule FORTRAN_DEP_HACK
             else:
                 exe_arr = exe.get_command()
             base_args = generator.get_arglist()
+            extra_dependencies = [os.path.join(self.build_to_src, i) for i in genlist.extra_depends]
             for i in range(len(infilelist)):
                 if len(generator.outputs) == 1:
                     sole_output = os.path.join(self.get_target_private_dir(target), outfilelist[i])
@@ -1008,6 +1027,8 @@ rule FORTRAN_DEP_HACK
                         for x in args]
                 cmdlist = exe_arr + args
                 elem = NinjaBuildElement(outfiles, 'CUSTOM_COMMAND', infilename)
+                if len(extra_dependencies) > 0:
+                    elem.add_dep(extra_dependencies)
                 elem.add_item('DESC', 'Generating $out')
                 if isinstance(exe, build.BuildTarget):
                     elem.add_dep(self.get_target_filename(exe))
@@ -1030,7 +1051,7 @@ rule FORTRAN_DEP_HACK
             # but those are really rare. I hope.
             if not compiler.can_compile(s):
                 continue
-            for line in open(os.path.join(self.environment.get_source_dir(), target.subdir, s)):
+            for line in open(os.path.join(self.environment.get_source_dir(), s.subdir, s.fname)):
                 modmatch = modre.match(line)
                 if modmatch is not None:
                     modname = modmatch.group(1)
@@ -1064,7 +1085,7 @@ rule FORTRAN_DEP_HACK
                 # Check if a source uses a module it exports itself.
                 # Potential bug if multiple targets have a file with
                 # the same name.
-                if mod_source_file == os.path.split(src)[1]:
+                if mod_source_file.fname == os.path.split(src)[1]:
                     continue
                 mod_name = compiler.module_name_to_filename(usematch.group(1))
                 mod_files.append(os.path.join(dirname, mod_name))
@@ -1098,10 +1119,15 @@ rule FORTRAN_DEP_HACK
                 rel_src = os.path.join(self.get_target_private_dir(target), src)
                 abs_src = os.path.join(self.environment.get_source_dir(), rel_src)
         else:
-            rel_src = os.path.join(self.build_to_src, target.get_source_subdir(), src)
+            if isinstance(src, File):
+                rel_src = src.rel_to_builddir(self.build_to_src)
+            else:
+                raise build.InvalidArguments('Invalid source type.')
             abs_src = os.path.join(self.environment.get_build_dir(), rel_src)
         if isinstance(src, RawFilename):
             src_filename = src.fname
+        elif isinstance(src, File):
+            src_filename = src.fname
         elif os.path.isabs(src):
             src_filename = os.path.basename(src)
         else:
