@@ -16,14 +16,27 @@
 
 from glob import glob
 import os, subprocess, shutil, sys, platform, signal
+from io import StringIO
+import sys
 import environment
 import mesonlib
+import mlog
+import meson, meson_test
 import argparse
 import xml.etree.ElementTree as ET
 import time
 
 from meson import backendlist
 
+class TestResult:
+    def __init__(self, msg, stdo, stde, conftime=0, buildtime=0, testtime=0):
+        self.msg = msg
+        self.stdo = stdo
+        self.stde = stde
+        self.conftime = conftime
+        self.buildtime = buildtime
+        self.testtime = testtime
+
 passing_tests = 0
 failing_tests = 0
 skipped_tests = 0
@@ -135,47 +148,76 @@ def log_text_file(logfile, testdir, msg, stdo, stde):
     if stop:
         raise StopException()
 
+def run_configure_inprocess(commandlist):
+    old_stdout = sys.stdout
+    sys.stdout = mystdout = StringIO()
+    old_stderr = sys.stderr
+    sys.stderr = mystderr = StringIO()
+    returncode = meson.run(commandlist)
+    sys.stdout = old_stdout
+    sys.stderr = old_stderr
+    return (returncode, mystdout.getvalue(), mystderr.getvalue())
+
+def run_test_inprocess(testdir):
+    old_stdout = sys.stdout
+    sys.stdout = mystdout = StringIO()
+    old_stderr = sys.stderr
+    sys.stderr = mystderr = StringIO()
+    old_cwd = os.getcwd()
+    os.chdir(testdir)
+    returncode = meson_test.run(['meson-private/meson_test_setup.dat'])
+    sys.stdout = old_stdout
+    sys.stderr = old_stderr
+    os.chdir(old_cwd)
+    return (returncode, mystdout.getvalue(), mystderr.getvalue())
+
+
 def run_test(testdir, should_succeed):
     global compile_commands
+    mlog.shutdown() # Close the log file because otherwise Windows wets itself.
     shutil.rmtree(test_build_dir)
     shutil.rmtree(install_dir)
     os.mkdir(test_build_dir)
     os.mkdir(install_dir)
     print('Running test: ' + testdir)
-    gen_command = [sys.executable, meson_command, '--prefix', '/usr', '--libdir', 'lib', testdir, test_build_dir]\
+    gen_start = time.time()
+    gen_command = [meson_command, '--prefix', '/usr', '--libdir', 'lib', testdir, test_build_dir]\
         + unity_flags + backend_flags
-    p = subprocess.Popen(gen_command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
-    (stdo, stde) = p.communicate()
-    stdo = stdo.decode('utf-8')
-    stde = stde.decode('utf-8')
+    (returncode, stdo, stde) = run_configure_inprocess(gen_command)
+    gen_time = time.time() - gen_start
     if not should_succeed:
-        if p.returncode != 0:
-            return ('', stdo, stde)
-        return ('Test that should have failed succeeded', stdo, stde)
-    if p.returncode != 0:
-        return ('Generating the build system failed.', stdo, stde)
+        if returncode != 0:
+            return TestResult('', stdo, stde, gen_time)
+        return TestResult('Test that should have failed succeeded', stdo, stde, gen_time)
+    if returncode != 0:
+        return TestResult('Generating the build system failed.', stdo, stde, gen_time)
     if 'msbuild' in compile_commands[0]:
         sln_name = glob(os.path.join(test_build_dir, '*.sln'))[0]
         comp = compile_commands + [os.path.split(sln_name)[-1]]
     else:
         comp = compile_commands
+    build_start = time.time()
     pc = subprocess.Popen(comp, cwd=test_build_dir,
                           stdout=subprocess.PIPE, stderr=subprocess.PIPE)
     (o, e) = pc.communicate()
+    build_time = time.time() - build_start
     stdo += o.decode('utf-8')
     stde += e.decode('utf-8')
     if pc.returncode != 0:
-        return ('Compiling source code failed.', stdo, stde)
-    pt = subprocess.Popen(test_commands, cwd=test_build_dir,
-                          stdout=subprocess.PIPE, stderr=subprocess.PIPE)
-    (o, e) = pt.communicate()
-    stdo += o.decode('utf-8')
-    stde += e.decode('utf-8')
-    if pt.returncode != 0:
-        return ('Running unit tests failed.', stdo, stde)
+        return TestResult('Compiling source code failed.', stdo, stde, gen_time, build_time)
+    test_start = time.time()
+    # Note that we don't test that running e.g. 'ninja test' actually
+    # works. One hopes that this is a common enough happening that
+    # it is picked up immediately on development.
+    (returncode, tstdo, tstde) = run_test_inprocess(test_build_dir)
+    test_time = time.time() - test_start
+    stdo += tstdo
+    stde += tstde
+    if returncode != 0:
+        return TestResult('Running unit tests failed.', stdo, stde, gen_time, build_time, test_time)
     if len(install_commands) == 0:
         print("Skipping install test")
-        return ('', '', '')
+        return TestResult('', '', '', gen_time, build_time, test_time)
     else:
         env = os.environ.copy()
         env['DESTDIR'] = install_dir
@@ -185,8 +227,8 @@ def run_test(testdir, should_succeed):
         stdo += o.decode('utf-8')
         stde += e.decode('utf-8')
         if pi.returncode != 0:
-            return ('Running install failed.', stdo, stde)
-        return (validate_install(testdir, install_dir), stdo, stde)
+            return TestResult('Running install failed.', stdo, stde, gen_time, build_time, test_time)
+        return TestResult(validate_install(testdir, install_dir), stdo, stde, gen_time, build_time, test_time)
 
 def gather_tests(testdir):
     tests = [t.replace('\\', '/').split('/', 2)[2] for t in glob(os.path.join(testdir, '*'))]
@@ -215,8 +257,11 @@ def detect_tests_to_run():
 
 def run_tests():
     all_tests = detect_tests_to_run()
-    logfile = open('meson-test-run.txt', 'w')
+    logfile = open('meson-test-run.txt', 'w', encoding="utf_8")
     junit_root = ET.Element('testsuites')
+    conf_time = 0
+    build_time = 0
+    test_time = 0
     try:
         os.mkdir(test_build_dir)
     except OSError:
@@ -245,18 +290,24 @@ def run_tests():
                 skipped_tests += 1
             else:
                 ts = time.time()
-                (msg, stdo, stde) = run_test(t, name != 'failing')
+                result = run_test(t, name != 'failing')
                 te = time.time()
-                log_text_file(logfile, t, msg, stdo, stde)
+                conf_time += result.conftime
+                build_time += result.buildtime
+                test_time += result.testtime
+                log_text_file(logfile, t, result.msg, result.stdo, result.stde)
                 current_test = ET.SubElement(current_suite, 'testcase', {'name' : testname,
                                                                          'classname' : name,
                                                                          'time' : '%.3f' % (te - ts)})
-                if msg != '':
-                    ET.SubElement(current_test, 'failure', {'message' : msg})
+                if result.msg != '':
+                    ET.SubElement(current_test, 'failure', {'message' : result.msg})
                 stdoel = ET.SubElement(current_test, 'system-out')
-                stdoel.text = stdo
+                stdoel.text = result.stdo
                 stdeel = ET.SubElement(current_test, 'system-err')
-                stdeel.text = stde
+                stdeel.text = result.stde
+    print("\nTotal configuration time: %.2fs" % conf_time)
+    print("Total build time: %.2fs" % build_time)
+    print("Total test time: %.2fs" % test_time)
     ET.ElementTree(element=junit_root).write('meson-test-run.xml', xml_declaration=True, encoding='UTF-8')
 
 def check_file(fname):
@@ -288,7 +339,13 @@ def generate_prebuilt_object():
             objectfile = objectbase + 'obj'
         else:
             objectfile = objectbase + 'o'
-        cmd = ['cc', '-c', source, '-o', objectfile]
+        if shutil.which('cc'):
+            cmd = 'cc'
+        elif shutil.which('gcc'):
+            cmd = 'gcc'
+        else:
+            raise RuntimeError("Could not find C compiler.")
+        cmd = [cmd, '-c', source, '-o', objectfile]
     subprocess.check_call(cmd)
     return objectfile
 
