@@ -16,6 +16,7 @@ import os, sys
 import backends, build
 import xml.etree.ElementTree as ET
 import xml.dom.minidom
+from coredata import MesonException
 
 class Vs2010Backend(backends.Backend):
     def __init__(self, build, interp):
@@ -143,7 +144,11 @@ class Vs2010Backend(backends.Backend):
     def target_to_build_root(self, target):
         if target.subdir == '':
             return ''
-        return '/'.join(['..']*(len(os.path.split(target.subdir))-1))
+
+        directories = os.path.split(target.subdir)
+        directories = list(filter(bool,directories)) #Filter out empty strings
+
+        return '/'.join(['..']*len(directories))
 
     def special_quote(self, arr):
         return ['&quot;%s&quot;' % i for i in arr]
@@ -217,6 +222,10 @@ class Vs2010Backend(backends.Backend):
         opt = ET.SubElement(clconf, 'Optimization')
         opt.text = 'disabled'
         inc_dirs = [proj_to_src_dir, self.get_target_private_dir(target)]
+        cur_dir = target.subdir
+        if cur_dir == '':
+            cur_dir= '.'
+        inc_dirs.append(cur_dir)
         extra_args = []
         # SUCKS, VS can not handle per-language type flags, so just use
         # them all.
@@ -283,15 +292,15 @@ class Vs2010Backend(backends.Backend):
         if len(headers) + len(gen_hdrs) > 0:
             inc_hdrs = ET.SubElement(root, 'ItemGroup')
             for h in headers:
-                relpath = os.path.join(proj_to_src_dir, h)
+                relpath = h.rel_to_builddir(proj_to_src_root)
                 ET.SubElement(inc_hdrs, 'CLInclude', Include=relpath)
             for h in gen_hdrs:
-                relpath = self.relpath(h, target.subdir)
+                relpath = h.rel_to_builddir(proj_to_src_root)
                 ET.SubElement(inc_hdrs, 'CLInclude', Include = relpath)
         if len(sources) + len(gen_src) > 0:
             inc_src = ET.SubElement(root, 'ItemGroup')
             for s in sources:
-                relpath = os.path.join(proj_to_src_dir, s)
+                relpath = s.rel_to_builddir(proj_to_src_root)
                 ET.SubElement(inc_src, 'CLCompile', Include=relpath)
             for s in gen_src:
                 relpath =  self.relpath(s, target.subdir)
